#!/usr/bin/env python

"""
Simplify AST structures for easier type propagation and analysis.

Copyright (C) 2006 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from compiler.visitor import ASTVisitor
import compiler.ast
from simplified import *

class Simplifier(ASTVisitor):

    """
    A simplifying visitor for AST nodes.

    Covered: AssAttr, AssList, AssName, AssTuple, Assign, AugAssign, Break,
             CallFunc, Class, Const, Continue, Dict, Discard, For, Function,
             Getattr, Global, If, Keyword, Lambda, List, Module, Name, Pass,
             Return, Stmt, TryExcept, TryFinally, Tuple, While.

    Missing: Add, And, Assert, Backquote, Bitand, Bitor, Bitxor, Compare,
             Decorators, Div, Ellipsis, Exec, FloorDiv, From, Import, Invert,
             LeftShift, ListComp, ListCompFor, ListCompIf, Mod, Mul, Not, Or,
             Power, Print, Printnl, Raise, RightShift, Slice, Sliceobj, Sub,
             Subscript, UnaryAdd, UnarySub, Yield.
    """

    def __init__(self):
        ASTVisitor.__init__(self)
        self.result = None              # The resulting tree.
        self.subprograms = []           # Subprograms outside the tree.
        self.current_subprograms = []   # Current subprograms being processed.

    # Generic visitor methods.

    def default(self, node, *args):
        raise ValueError, node.__class__

    def dispatch(self, node, *args):
        return ASTVisitor.dispatch(self, node, *args)

    def dispatches(self, nodes, *args):
        results = []
        for node in nodes:
            results.append(self.dispatch(node, *args))
        return results

    # Placeholder or deletion transformations.

    def visitStmt(self, stmt):
        return self.dispatches(stmt.nodes)

    def visitPass(self, pass_):
        return Pass(pass_)

    def visitDiscard(self, discard):
        return self.dispatch(discard.expr)

    # Relatively trivial transformations.

    def visitModule(self, module):
        self.result = Module(module)
        self.result.code = self.dispatch(module.node)
        return self.result

    def visitClass(self, class_):
        result = Class(class_, name=class_.name, bases=class_.bases)
        result.code = self.dispatch(class_.code)
        return result

    def visitGetattr(self, getattr):
        result = LoadAttr(getattr, name=getattr.attrname)
        result.expr = self.dispatch(getattr.expr)
        return result

    def visitKeyword(self, keyword):
        result = Keyword(keyword, name=keyword.name)
        result.expr = self.dispatch(keyword.expr)
        return result

    def visitGlobal(self, global_):
        result = Global(global_, names=global_.names)
        return result

    def visitName(self, name):
        result = LoadName(name, name=name.name)
        return result

    def visitConst(self, const):
        result = LoadConst(const, value=const.value)
        return result

    def visitReturn(self, return_):
        result = Return(return_)
        result.expr = self.dispatch(return_.value)
        return result

    def visitBreak(self, break_):
        result = Return(break_)
        return result

    def visitContinue(self, continue_):
        result = Invoke(continue_, same_frame=1, star=None, dstar=None, args=[])
        result.expr = LoadRef(ref=self.current_subprograms[-1])
        return result

    def visitIf(self, if_):
        result = If(if_, else_=[])
        tests = []
        for compare, stmt in if_.tests:
            # Produce something like...
            # expr.__true__() ? body
            test = Conditional(else_=[], test=Invoke(
                expr=LoadAttr(expr=self.dispatch(compare), name="__true__"),
                params=[], star=None, dstar=None))
            test.body = self.dispatch(stmt)
            tests.append(test)
        result.tests = tests
        if if_.else_ is not None:
            result.else_ = self.dispatch(if_.else_)
        return result

    def _visitBuiltin(self, builtin, name):
        result = Invoke(builtin, expr=LoadName(name=name))
        result.args = self.dispatches(builtin.nodes)
        return result

    def visitTuple(self, tuple):
        return self._visitBuiltin(tuple, "Tuple")

    def visitList(self, list):
        return self._visitBuiltin(list, "List")

    def visitDict(self, dict):
        result = Invoke(expr=LoadName(name="Dict"))
        args = []
        for key, value in dict.items:
            tuple = Invoke(expr=LoadName(name="Tuple"), star=None, dstar=None)
            tuple.args = [self.dispatch(key), self.dispatch(value)]
            args.append(tuple)
        result.args = args
        return result

    # Logical operators.

    def visitAnd(self, and_):
        result = And(and_)
        nodes = []
        for node in and_.nodes:
            nodes.append(Invoke(expr=LoadAttr(expr=self.dispatch(node), name="__true__"),
                params=[], star=None, dstar=None))
        result.nodes = nodes
        return result

    def visitOr(self, or_):
        result = Or(or_)
        nodes = []
        for node in or_.nodes:
            nodes.append(Invoke(expr=LoadAttr(expr=self.dispatch(node), name="__true__"),
                params=[], star=None, dstar=None))
        result.nodes = nodes
        return result

    def visitNot(self, not_):
        result = Not(not_, expr=Invoke(expr=LoadAttr(expr=self.dispatch(not_.expr), name="__true__"),
            params=[], star=None, dstar=None))
        return result

    # Assignments.

    augassign_methods = {
        "+=" : "__iadd__", "-=" : "__isub__", "*=" : "__imul__", "/=" : "__idiv__"
        }

    def visitAugAssign(self, augassign):
        result = Assign(augassign)
        expr = self.dispatch(augassign.expr)

        # Simple augmented assignment: name += expr

        if isinstance(augassign.node, compiler.ast.Name):
            node = self.dispatch(augassign.node)
            get_incremented = StoreTemp(
                expr=Invoke(expr=LoadAttr(expr=node, name=self.augassign_methods[augassign.op]), args=[expr])
                )
            store = StoreName(expr=LoadTemp(), name=augassign.node.name)
            result.code = [get_incremented, store, ReleaseTemp()]

        # Complicated augmented assignment: expr.attr += expr

        elif isinstance(augassign.node, compiler.ast.Getattr):
            store_expr = StoreTemp(index=1, expr=self.dispatch(augassign.node.expr))
            node_attr = LoadAttr(expr=LoadTemp(index=1), name=augassign.node.attrname)
            get_incremented = StoreTemp(
                expr=Invoke(expr=LoadAttr(expr=node_attr, name=self.augassign_methods[augassign.op]), args=[expr])
                )
            store = StoreAttr(expr=LoadTemp(), lvalue=LoadTemp(index=1), name=augassign.node.attrname)
            result.code = [store_expr, get_incremented, store, ReleaseTemp(index=1), ReleaseTemp()]

        else:
            raise NotImplementedError, augassign.node.__class__

        return result

    def visitAssign(self, assign):
        result = Assign(assign)
        store = StoreTemp(expr=self.dispatch(assign.expr))
        release = ReleaseTemp()
        result.code = [store] + self.dispatches(assign.nodes, 0) + [release]
        return result

    def visitAssList(self, asslist, in_sequence=0):
        if not in_sequence:
            expr = LoadTemp(asslist)
        else:
            expr = Invoke(asslist, expr=LoadAttr(expr=LoadTemp(), name="next"))
        result = Assign(asslist)
        store = StoreTemp(expr=Invoke(expr=LoadAttr(name="__iter__", expr=expr)))
        release = ReleaseTemp()
        result.code = [store] + self.dispatches(asslist.nodes, 1) + [release]
        return result

    visitAssTuple = visitAssList

    def _visitAssNameOrAttr(self, node, in_sequence):
        if not in_sequence:
            return LoadTemp(node)
        else:
            return Invoke(node, expr=LoadAttr(expr=LoadTemp(), name="next"))

    def visitAssName(self, assname, in_sequence=0):
        expr = self._visitAssNameOrAttr(assname, in_sequence)
        result = StoreName(assname, name=assname.name, expr=expr)
        return result

    def visitAssAttr(self, assattr, in_sequence=0):
        expr = self._visitAssNameOrAttr(assattr, in_sequence)
        lvalue = self.dispatch(assattr.expr)
        result = StoreAttr(assattr, name=assattr.attrname, lvalue=lvalue, expr=expr)
        return result

    # Invocation and subprogram transformations.

    def _visitFunction(self, function, subprogram):
        if function.flags & 4 != 0: has_star = 1
        else: has_star = 0
        if function.flags & 8 != 0: has_dstar = 1
        else: has_dstar = 0
        ndefaults = len(function.defaults)
        npositional = len(function.argnames) - has_star - has_dstar
        if has_star: star = function.argnames[npositional]
        else: star = None
        if has_dstar: dstar = function.argnames[npositional + has_star]
        else: dstar = None

        params = []
        for i in range(0, npositional - ndefaults):
            params.append((function.argnames[i], None))

        # NOTE: Fix/process defaults.

        for i in range(0, ndefaults):
            default = function.defaults[i]
            if default is not None:
                params.append((function.argnames[npositional - ndefaults + i], self.dispatch(default)))
            else:
                params.append((function.argnames[npositional - ndefaults + i], default))

        subprogram.params = params
        subprogram.star = star
        subprogram.dstar = dstar
        self.subprograms.append(subprogram)

    def visitFunction(self, function):

        # Make a subprogram for the function and record it outside the main
        # tree.

        subprogram = Subprogram(function, name=function.name, star=None, dstar=None)
        self.current_subprograms.append(subprogram)
        subprogram.code = self.dispatch(function.code)
        self.current_subprograms.pop()
        self._visitFunction(function, subprogram)

        # Make a definition of the function associating it with a name.

        result = Assign(function)
        load = LoadRef(ref=subprogram)
        store = StoreName(name=function.name)
        result.code = [load, store]
        return result

    def visitLambda(self, lambda_):

        # Make a subprogram for the function and record it outside the main
        # tree.

        subprogram = Subprogram(lambda_, name=hex(id(lambda_)), star=None, dstar=None)
        self.current_subprograms.append(subprogram)
        subprogram.code = [Return(expr=self.dispatch(lambda_.code))]
        self.current_subprograms.pop()
        self._visitFunction(lambda_, subprogram)

        # Get the subprogram reference to the lambda.

        return LoadRef(ref=subprogram)

    def visitCallFunc(self, callfunc):
        result = Invoke(callfunc, same_frame=0, star=None, dstar=None)
        result.args = self.dispatches(callfunc.args)
        if callfunc.star_args is not None:
            result.star = self.dispatch(callfunc.star_args)
        if callfunc.dstar_args is not None:
            result.dstar = self.dispatch(callfunc.dstar_args)
        result.expr = self.dispatch(callfunc.node)
        return result

    def visitWhile(self, while_):

        # Make a subprogram for the block and record it outside the main tree.

        subprogram = Subprogram(while_, name=hex(id(while_)), acquire_locals=1, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # Include a conditional statement in the subprogram.

        test = Conditional(else_=[])
        test.test = Invoke(expr=LoadAttr(expr=self.dispatch(while_.test), name="__true__"),
            params=[], star=None, dstar=None)

        # Inside the conditional, add a recursive invocation to the subprogram
        # if the test condition was satisfied.

        continuation = Invoke(same_frame=1, star=None, dstar=None, args=[])
        continuation.expr = LoadRef(ref=subprogram)
        test.body = self.dispatch(while_.body) + [continuation]
        if while_.else_ is not None:
            test.else_ = self.dispatch(while_.else_)
        subprogram.code = [test]

        self.current_subprograms.pop()
        self.subprograms.append(subprogram)

        # Make an invocation of the subprogram.

        result = Invoke(while_, same_frame=1, star=None, dstar=None, args=[])
        result.expr = LoadRef(ref=subprogram)
        return result

    def visitFor(self, for_):

        # Make a subprogram for the block and record it outside the main tree.

        subprogram = Subprogram(for_, name=hex(id(for_)), acquire_locals=1, params=[], star=None, dstar=None)
        self.current_subprograms.append(subprogram)

        # Wrap the assignment in a try...except statement.

        try_except = Try(body=[], handlers=[], else_=[], finally_=[])
        except_spec = Invoke(expr=LoadName(name="Tuple"), params=[LoadName(name="StopIteration")])
        stopiteration = Except(spec=except_spec)
        stopiteration.code = self.dispatch(for_.else_)
        try_except.handlers = [stopiteration]

        assign = Assign()
        assign.code = [
            StoreTemp(expr=Invoke(expr=LoadAttr(expr=LoadTemp(), name="next"))),
            self.dispatch(for_.assign),
            ReleaseTemp()
            ]

        # Inside the conditional, add a recursive invocation to the subprogram
        # if the test condition was satisfied.

        continuation = Invoke(same_frame=1, star=None, dstar=None, args=[])
        continuation.expr = LoadRef(ref=subprogram)
        try_except.body = [assign] + self.dispatch(for_.body) + [continuation]
        subprogram.code = [try_except]

        self.subprograms.append(subprogram)
        self.current_subprograms.pop()

        # Obtain an iterator for the sequence involved.
        # Then, make an invocation of the subprogram.

        result = Assign(for_)
        result.code = [
            StoreTemp(expr=Invoke(expr=LoadAttr(name="__iter__", expr=self.dispatch(for_.list)))),
            Invoke(expr=LoadRef(ref=subprogram), same_frame=1, star=None, dstar=None, args=[]),
            ReleaseTemp()
            ]
        return result

    # Exception node transformations.

    def visitTryExcept(self, tryexcept):
        result = Try(tryexcept, body=[], handlers=[], else_=[], finally_=[])
        if tryexcept.body is not None:
            result.body = self.dispatch(tryexcept.body)
        if tryexcept.else_ is not None:
            result.else_ = self.dispatch(tryexcept.else_)
        handlers = []
        for spec, assign, stmt in tryexcept.handlers:
            get_exc = Assign()
            get_exc.code = [StoreTemp(expr=LoadExc()), self.dispatch(assign), ReleaseTemp()]
            handler = Except(spec=self.dispatch(spec))
            handler.code = [get_exc] + self.dispatch(stmt)
            handlers.append(handler)
        result.handlers = handlers
        return result

    def visitTryFinally(self, tryfinally):
        result = Try(tryfinally, body=[], handlers=[], else_=[], finally_=[])
        if tryfinally.body is not None:
            result.body = self.dispatch(tryfinally.body)
        if tryfinally.final is not None:
            result.finally_ = self.dispatch(tryfinally.final)
        return result

# vim: tabstop=4 expandtab shiftwidth=4
